<?xml version="1.0" encoding="UTF-8"?><jobInfo
   xmlns="http://www.force.com/2009/06/asyncapi/dataload">
 <id>{id}</id>
 <operation>upsert</operation>
 <object>{object}</object>
 <createdById>{creator}</createdById>
 <createdDate>{cdate}</createdDate>
 <systemModstamp>{smts}</systemModstamp>
 <state>{state:Open}</state>
 <externalIdFieldName>{idField:Id}</externalIdFieldName>
 <concurrencyMode>Parallel</concurrencyMode>
 <contentType>CSV</contentType>
 <numberBatchesQueued>{qdbatches:0}</numberBatchesQueued>
 <numberBatchesInProgress>{numbatchip:0}</numberBatchesInProgress>
 <numberBatchesCompleted>{numbatchcomp:0}</numberBatchesCompleted>
 <numberBatchesFailed>{numbatchfail:0}</numberBatchesFailed>
 <numberBatchesTotal>{numbatchtotal:1}</numberBatchesTotal>
 <numberRecordsProcessed>{numrecsprocc:0}</numberRecordsProcessed>
 <numberRetries>{numretries:0}</numberRetries>
 <apiVersion>vxx.0</apiVersion>
 <numberRecordsFailed>{numrecsfail:0}</numberRecordsFailed>
 <totalProcessingTime>{tpt:0}</totalProcessingTime>
 <apiActiveProcessingTime>{apipt:0}</apiActiveProcessingTime>
 <apexProcessingTime>{apext:0}</apexProcessingTime>
</jobInfo>