<?xml version="1.0" encoding="UTF-8"?><batchInfoList
   xmlns="http://www.force.com/2009/06/asyncapi/dataload">
 <batchInfo>
  <id>{id}</id>
  <jobId>{jobId}</jobId>
  <state>{state}</state>
  <createdDate>{cdate}</createdDate>
  <systemModstamp>{smts}</systemModstamp>
  <numberRecordsProcessed>{numrecsprocc:0}</numberRecordsProcessed>
  <numberRecordsFailed>{numrecsfail:0}</numberRecordsFailed>
  <totalProcessingTime>{tpt:0}</totalProcessingTime>
  <apiActiveProcessingTime>{apipt:0}</apiActiveProcessingTime>
  <apexProcessingTime>{apext:0}</apexProcessingTime>
 </batchInfo>
</batchInfoList>